#include "ext.h"
#include "z_dsp.h" 

void *localmax_class;

typedef struct _localmax
{
    t_pxobject	x_obj;
    
    double 		hold_time_ms;				// in ms
    long 		hold_time_s;				// in samples
    long 		sample_count;				// sample increment for hold time
    
    double		peaks[20];					// keep last peak every hold_time_s samples
    long		sdiffpeaks[20];				// sign of differential of peaks
    long		diffsdiffpeaks[20];			// diff of above vector
    double		out_peak[20];				// so we know if we need ti output~ this peak 
    
    double		local_peak;					// local peak in last hold time duration

    long		gate_state;					// 0 is close, 1 let input~ pass to output~
    double		seuil;						// if value is above seuil then close localmax internal gate during given time

    double 		s_rate;
    long		sigvs;

    // void 		*l_out;					// double outlet for peak
    // void 		*l_out2;				// localmax flag (1=limit, 0 = no limit)

} t_localmax, *t_localmax_ptr;

t_int *localmax_perform(t_int *w);
void *localmax_new(double hold_time_ms,double seuil);
void localmax_dsp(t_localmax *x, t_signal **sp, short *count);
void localmax_assist(t_localmax *x, void *b, long m, long a, char *s);
void localmax_free(t_localmax *x);
void localmax_debug(t_localmax *x);
void localmax_hold_time_ms(t_localmax *x, double time);
void localmax_seuil(t_localmax *x, double seuil);
void rotate_d(double *a,long n);
void rotate_l(long *a,long n);

void main(void)
{
    setup((t_messlist **) &localmax_class, (method) localmax_new, (method)localmax_free, (short)sizeof(t_localmax), 0L,A_DEFFLOAT,A_DEFFLOAT,0);
    addmess((method)localmax_dsp, "dsp", A_CANT, 0);
    addmess((method)localmax_hold_time_ms,"hold_time_ms",A_FLOAT,0);
    addmess((method)localmax_seuil,"seuil",A_FLOAT,0);
	addmess((method)localmax_assist,"assist",A_CANT,0);
	addmess((method)localmax_debug,"debug",0);
    post("gn.localmax~ v0.4 UB 6.V.2007 @ Gilbert Nouno ");
    dsp_initclass(); // set object class to work with MSP
    rescopy('STR#',4100);
}


void localmax_assist(t_localmax *x, void *b, long m, long a, char *s)
{
	if (m == ASSIST_OUTLET) { // outlet
		sprintf(s,"localmax~ "); // titre assist
		switch (a) {	
		case 0:
			sprintf(s,"local max value if found within time window else 0.");
			break;
		case 1:
			sprintf(s,"not used here because one outlet");
			break;
		}
		}
	else { // inlets
		switch (a) {
		case 0:
			sprintf(s,"input signal");
			break;
		case 1:
			sprintf(s,"not used here beacause one inlet");
			break;
		}
	}
}

void *localmax_new(double hold_time_ms,double seuil)
{
    t_localmax *x = (t_localmax *)newobject(localmax_class);
    dsp_setup((t_pxobject *)x,1); // localmax has one input signal
    // x->l_out 	= outlet_new((t_pxobject *)x, 0L); 
    outlet_new((t_pxobject *)x, "signal");
   
    x->x_obj.z_misc	=	Z_NO_INPLACE; // in & out independants    
    
    x->s_rate		=	44100;			// peut-on demander le sample rate ici ???????????????
    x->hold_time_ms	=	(hold_time_ms <= 0.) ? 5. : hold_time_ms;
	x->hold_time_s	=	round(x->hold_time_ms*0.001*x->s_rate);	
	x->seuil		=	(seuil <= 0.) ? 0. : seuil;
    x->sample_count	=	0;
    x->local_peak	=	0;
    return (x);
}

void localmax_free(t_localmax *x) {
	dsp_free((t_pxobject *)x);
}

void localmax_dsp(t_localmax *x, t_signal **sp, short *count)
{
	x->s_rate	=	sys_getsr();
	x->sigvs	=	sys_getblksize(); // ne marche pas pourquoi ???
	dsp_add(localmax_perform, 4, x, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n);
}

t_int *localmax_perform(t_int *w)	{
	float		*in, *out, *in_copy, *out_copy;
	double		possible_peak;
	double		seuil,   local_peak;
	long		hold_time_s;
	long		sample_count;
	
	long		gate_state,n,n_copy;
	short		click = 0;

	t_localmax *x=	(t_localmax *)(w[1])	;
	in 			= 	(float *)(w[2])		;		// vecteur d'entre
	out 		= 	(float *)(w[3])		;		// vecteur de sortie
	n 			= 	w[4]				;		// sigvs
	
	local_peak		= 	x->local_peak	;
	seuil 			= 	x->seuil		;
	sample_count	=	x->sample_count	;
	gate_state		=	x->gate_state	;
	hold_time_s		=	x->hold_time_s	;
	
	in_copy		=	in					;
	out_copy	=	out					;
	n_copy		=	n					;
	
	while (n--) {  // check si on doit sortir un local max, si oui alors le faire en debut de vecteur hold_time_ms
				long	i;
				sample_count++;
				if ( *in > local_peak) local_peak=*in;
				if (sample_count > hold_time_s) { 
										rotate_d(x->peaks,20);
										rotate_d(x->out_peak,20);
										x->peaks[19]= local_peak;
										x->out_peak[19]= 0;
										click =  ( x->out_peak[1] > seuil) ? 1 : 0 ;
										local_peak = 0;
										sample_count = 0;
										  
										// on calcule alors les sdiff et diffsdiff
										// calcul des diffs   .... faire des rotations ..... mieux
										x->sdiffpeaks[0] = 0;
										x->diffsdiffpeaks[0] = 0;
										for(i=1;i< 20 ;i++) { x->sdiffpeaks[i]=  ((x->peaks[i] - x->peaks[i-1]) >= 0)? 1:-1 ; } 
										for(i=1;i< 20 ;i++) { x->diffsdiffpeaks[i]=  x->sdiffpeaks[i] - x->sdiffpeaks[i-1]; }
	
										// les candidats maximum sont ceux pour lesquels diffsdiffpeaks est < 0
										// on regarde le 5e echantillon avec une fenetre de 4  gauche et  droite
										x->out_peak[5] = 0;
										if ( x->diffsdiffpeaks[5] <0) { possible_peak = x->peaks[4];
																		x->out_peak[5] = possible_peak;
																		for(i=1;i<= 9 ;i++) { if (x->peaks[i]>possible_peak) x->out_peak[5] = 0;}
										 								}
										  
										  }
										  
				if (click == 1)	{ 
					*out++ = x->out_peak[1];
					click = 0;
				} // output peak next ms 
				else *out++ = 0.;
				*in++;	
	}
	
	// save stuffs
	x->sample_count	=	sample_count	;
	x->local_peak	=	local_peak	;
	
	return (w+5);
}

void localmax_hold_time_ms(t_localmax *x, double time)	{
	x->hold_time_ms	= 	(time <= 0.) ? 100. : time;
	x->hold_time_s	=	round(x->hold_time_ms*0.001*x->s_rate);
    // post("hold_time_s =%f", x->hold_time_s);
}

void localmax_seuil(t_localmax *x, double seuil)	{
	x->seuil	= 	(seuil <= 0.) ? 0. : seuil;
   //  post("seuil =%f", x->seuil);
}

void localmax_debug(t_localmax *x)	{
	long	i;
	post("s_rate =%f", x->s_rate);
	post("sigvs  =%f", x->sigvs);
    post("sample_count  =%d", x->sample_count);
	post("hold_time_ms =%f", x->hold_time_ms);
    post("hold_time_s =%d", x->hold_time_s);
    post("seuil =%f", x->seuil);
    for(i=0;i<20;i++) { post("peaks[%ld]=%f ",i, x->peaks[i]); } 
    for(i=0;i<20;i++) { post("sdiffpeaks[%ld]=%ld ",i, x->sdiffpeaks[i]); }
    for(i=0;i<20;i++) { post("diffsdiffpeaks[%ld]=%ld ",i, x->diffsdiffpeaks[i]); }
    for(i=0;i<6;i++) { post("out_peak[%ld]=%f ",i, x->out_peak[i]); }
}
	
void rotate_d(double *a,long n) {
	long	i;
	for(i=0;i<(n-1);i++) { a[i]=a[i+1]; }   // rotate left, last one should be at index n after rotation
	}
	
void rotate_l(long *a,long n) {
	long	i;
	for(i=0;i<(n-1);i++) { a[i]=a[i+1]; }   // rotate left, last one should be at index n after rotation
	}

